"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable-next-line: no-empty
var EMPTY_FN = function () { };
var IS_UNWAIT_PROXY_SYMBOL = Symbol.for("IS_UNWAIT_PROXY_SYMBOL");
var UNWAIT_PROMISE_VALUE = Symbol.for("UNWAIT_PROMISE_VALUE");
var cache = new WeakMap();
function accessProperty(obj, propertyName) {
    if (!obj) {
        throw new TypeError("Cannot access property " + propertyName.toString() + " of " + obj);
    }
    return obj[propertyName];
}
function apply(target, thisArg, argArray) {
    if (!target || typeof target !== "function") {
        throw new TypeError(target + " is not a function");
    }
    return target.apply(thisArg, argArray);
}
function construct(Target, argArray) {
    if (!Target || !Target.prototype) {
        throw new TypeError(Target + " is not a constructor");
    }
    return new (Target.bind.apply(Target, __spreadArrays([void 0], argArray)))();
}
/**
 * Helper to wrap a promise and expose it's value synchronously when resolved
 */
function wrapPromise(value) {
    var promise = Promise.resolve(value);
    promise.then(function (res) {
        promise[UNWAIT_PROMISE_VALUE] = res;
    });
    return promise;
}
/**
 * Internal unwait implementation
 * @param target the target to proxy
 * @param parent Provides the correct `this` context to wrapped child methods
 */
function _unwait(target, parent) {
    if (cache.has(target)) {
        return cache.get(target);
    }
    var promise = wrapPromise(target);
    var asyncProxyHandler = {
        construct: function (_, argArray) {
            return _unwait(promise.then(function (res) { return construct(res, argArray); }));
        },
        get: function (_, key) {
            // provide a way to determine if a value is an unwait proxy
            if (key === IS_UNWAIT_PROXY_SYMBOL) {
                return true;
            }
            // promise methods and properties should remain unchanged
            if (key in promise) {
                var property = promise[key];
                return typeof property === "function"
                    ? property.bind(promise)
                    : property;
            }
            // prefer using resolved value to allow cache lookup
            return promise[UNWAIT_PROMISE_VALUE]
                ? _unwait(accessProperty(promise[UNWAIT_PROMISE_VALUE], key), promise)
                : _unwait(promise.then(function (res) { return accessProperty(res, key); }), promise);
        },
        apply: function (_, thisArg, argArray) {
            if (thisArg && thisArg[IS_UNWAIT_PROXY_SYMBOL]) {
                thisArg = undefined;
            }
            var result;
            // patch the thisArg to be the parent unless specified
            if (parent && !thisArg) {
                // prefer using resolved value to allow cache lookup
                result = promise[UNWAIT_PROMISE_VALUE]
                    ? parent.then(function (parentThis) {
                        return apply(promise[UNWAIT_PROMISE_VALUE], parentThis, argArray);
                    })
                    : Promise.all([promise, parent]).then(function (_a) {
                        var res = _a[0], parentThis = _a[1];
                        return apply(res, parentThis, argArray);
                    });
            }
            else {
                // prefer using resolved value to allow cache lookup
                result = promise[UNWAIT_PROMISE_VALUE]
                    ? apply(promise[UNWAIT_PROMISE_VALUE], thisArg, argArray)
                    : promise.then(function (res) { return apply(res, thisArg, argArray); });
            }
            return _unwait(result, promise);
        },
    };
    var proxy = new Proxy(EMPTY_FN, asyncProxyHandler);
    promise = wrapPromise(promise.then(function (res) {
        if (typeof res === "object" || typeof res === "function") {
            cache.set(res, proxy);
        }
        return res;
    }));
    return proxy;
}
/**
 * Creates a proxy that defers all property and function access until it's target result has resolved.
 *
 * The target, and any of it's properties and method results can be called immediately,
 * and will return a proxied promise of the result when all promises in the chain resolve.
 * @param target The target to proxy.
 */
function unwait(target) {
    return _unwait(target);
}
exports.unwait = unwait;
exports.default = unwait;
